@@ -61,6 +61,7 @@
 #include "events/pong.h"
 #include "events/privmsg.h"
 #include "events/servlist.h"
+#include "events/stats.h"
 #include "events/wallops.h"
 #include "events/welcome.h"
 #include "events/whois.h"
@@ -126,22 +127,32 @@ static struct numeric_events_tag {
 	{ "005", "RPL_ISUPPORT",            NO_WINDOW,      0, event_serverFeatures },
 	{ "020", "",                        STATUS_WINDOW,  1, NULL },
 	{ "042", "",                        NO_WINDOW,      0, event_allaround_extract_remove_colon },
+	{ "211", "RPL_STATSLINKINFO",       ACTIVE_WINDOW,  1, NULL }, // XXX
+	{ "212", "RPL_STATSCOMMANDS",       ACTIVE_WINDOW,  1, NULL }, // XXX
+	{ "216", "RPL_STATSKLINE",          NO_WINDOW,      0, event_statskline },
+	{ "219", "RPL_ENDOFSTATS",          ACTIVE_WINDOW,  2, NULL },
 	{ "221", "RPL_UMODEIS",             NO_WINDOW,      0, event_userModeIs },
 	{ "232", "",                        ACTIVE_WINDOW,  1, NULL },
 	{ "234", "RPL_SERVLIST",            NO_WINDOW,      0, event_servlist },
 	{ "235", "RPL_SERVLISTEND",         NO_WINDOW,      0, event_servlistEnd },
+	{ "242", "RPL_STATSUPTIME",         ACTIVE_WINDOW,  1, NULL },
 	{ "250", "",                        STATUS_WINDOW,  1, NULL },
 	{ "251", "RPL_LUSERCLIENT",         STATUS_WINDOW,  1, NULL },
 	{ "252", "RPL_LUSEROP",             NO_WINDOW,      0, event_allaround_extract_remove_colon },
 	{ "253", "RPL_LUSERUNKNOWN",        NO_WINDOW,      0, event_allaround_extract_remove_colon },
 	{ "254", "RPL_LUSERCHANNELS",       NO_WINDOW,      0, event_allaround_extract_remove_colon },
 	{ "255", "RPL_LUSERME",             STATUS_WINDOW,  1, NULL },
+	{ "256", "RPL_ADMINME",             ACTIVE_WINDOW,  2, NULL },
+	{ "257", "RPL_ADMINLOC1",           ACTIVE_WINDOW,  1, NULL },
+	{ "258", "RPL_ADMINLOC2",           ACTIVE_WINDOW,  1, NULL },
+	{ "259", "RPL_ADMINEMAIL",          ACTIVE_WINDOW,  1, NULL },
 	{ "263", "RPL_TRYAGAIN",            ACTIVE_WINDOW,  2, NULL },
 	{ "265", "",                        NO_WINDOW,      0, event_local_and_global_users },
 	{ "266", "",                        NO_WINDOW,      0, event_local_and_global_users },
 	{ "275", "",                        NO_WINDOW,      0, event_whois_ssl },
 	{ "276", "",                        NO_WINDOW,      0, event_whois_cert },
 	{ "301", "RPL_AWAY",                NO_WINDOW,      0, event_whois_away },
+	{ "303", "RPL_ISON",                ACTIVE_WINDOW,  1, NULL },
 	{ "305", "RPL_UNAWAY",              NO_WINDOW,      0, event_unaway },
 	{ "306", "RPL_NOWAWAY",             NO_WINDOW,      0, event_nowAway },
 	{ "307", "",                        NO_WINDOW,      0, event_whois_service },
@@ -181,7 +192,9 @@ static struct numeric_events_tag {
 	{ "367", "RPL_BANLIST",             NO_WINDOW,      0, event_banlist },
 	{ "368", "RPL_ENDOFBANLIST",        NO_WINDOW,      0, event_eof_banlist },
 	{ "369", "RPL_ENDOFWHOWAS",         ACTIVE_WINDOW,  2, NULL },
+	{ "371", "RPL_INFO",                ACTIVE_WINDOW,  1, NULL },
 	{ "372", "RPL_MOTD",                NO_WINDOW,      0, event_motd },
+	{ "374", "RPL_ENDOFINFO",           ACTIVE_WINDOW,  1, NULL },
 	{ "375", "RPL_MOTDSTART",           NO_WINDOW,      0, event_motd },
 	{ "376", "RPL_ENDOFMOTD",           NO_WINDOW,      0, event_motd },
 	{ "378", "",                        NO_WINDOW,      0, event_whois_conn },
