@@ -477,12 +477,15 @@ static inline bool
 isInCirculationMode(const TAB_COMPLETION *tc)
 {
 	return (tc->isInCirculationModeForConnect ||
+	    tc->isInCirculationModeForCs ||
 	    tc->isInCirculationModeForHelp ||
 	    tc->isInCirculationModeForMsg ||
 	    tc->isInCirculationModeForNotice ||
+	    tc->isInCirculationModeForNs ||
 	    tc->isInCirculationModeForQuery ||
 	    tc->isInCirculationModeForSasl ||
 	    tc->isInCirculationModeForSettings ||
+	    tc->isInCirculationModeForSquery ||
 	    tc->isInCirculationModeForTheme ||
 	    tc->isInCirculationModeForTime ||
 	    tc->isInCirculationModeForVersion ||
