@@ -31,6 +31,8 @@
 
 #include "commands/connect.h"
 #include "commands/sasl.h"
+#include "commands/services.h"
+#include "commands/squery.h"
 #include "commands/theme.h"
 #include "commands/znc.h"
 
@@ -70,6 +72,13 @@ auto_complete_connect(volatile struct readline_session_context *ctx,
 	do_work(ctx, L"/connect ", s);
 }
 
+static void
+auto_complete_cs(volatile struct readline_session_context *ctx,
+    CSTRING s)
+{
+	do_work(ctx, L"/cs ", s);
+}
+
 static void
 auto_complete_help(volatile struct readline_session_context *ctx,
     CSTRING s)
@@ -91,6 +100,13 @@ auto_complete_notice(volatile struct readline_session_context *ctx,
 	do_work(ctx, L"/notice ", s);
 }
 
+static void
+auto_complete_ns(volatile struct readline_session_context *ctx,
+    CSTRING s)
+{
+	do_work(ctx, L"/ns ", s);
+}
+
 static void
 auto_complete_query(volatile struct readline_session_context *ctx,
     CSTRING s)
@@ -112,6 +128,13 @@ auto_complete_setting(volatile struct readline_session_context *ctx,
 	do_work(ctx, L"/set ", s);
 }
 
+static void
+auto_complete_squery(volatile struct readline_session_context *ctx,
+    CSTRING s)
+{
+	do_work(ctx, L"/squery ", s);
+}
+
 static void
 auto_complete_theme(volatile struct readline_session_context *ctx,
     CSTRING s)
@@ -234,20 +257,23 @@ readline_tab_comp_ctx_new(void)
 
 	BZERO(ctx.search_var, sizeof ctx.search_var);
 
-	ctx.isInCirculationModeForConnect	= false;
-	ctx.isInCirculationModeForHelp		= false;
-	ctx.isInCirculationModeForMsg		= false;
-	ctx.isInCirculationModeForNotice	= false;
-	ctx.isInCirculationModeForQuery		= false;
-	ctx.isInCirculationModeForSasl		= false;
-	ctx.isInCirculationModeForSettings	= false;
-	ctx.isInCirculationModeForTheme		= false;
-	ctx.isInCirculationModeForTime		= false;
-	ctx.isInCirculationModeForVersion	= false;
-	ctx.isInCirculationModeForWhois		= false;
-	ctx.isInCirculationModeForZncCmds	= false;
-	ctx.isInCirculationModeForCmds		= false;
-	ctx.isInCirculationModeForChanUsers	= false;
+	ctx.isInCirculationModeForConnect =
+	ctx.isInCirculationModeForCs =
+	ctx.isInCirculationModeForHelp =
+	ctx.isInCirculationModeForMsg =
+	ctx.isInCirculationModeForNotice =
+	ctx.isInCirculationModeForNs =
+	ctx.isInCirculationModeForQuery =
+	ctx.isInCirculationModeForSasl =
+	ctx.isInCirculationModeForSettings =
+	ctx.isInCirculationModeForSquery =
+	ctx.isInCirculationModeForTheme =
+	ctx.isInCirculationModeForTime =
+	ctx.isInCirculationModeForVersion =
+	ctx.isInCirculationModeForWhois =
+	ctx.isInCirculationModeForZncCmds =
+	ctx.isInCirculationModeForCmds =
+	ctx.isInCirculationModeForChanUsers = false;
 
 	ctx.matches = NULL;
 	ctx.elmt = NULL;
@@ -268,20 +294,23 @@ readline_tab_comp_ctx_reset(PTAB_COMPLETION ctx)
 	if (ctx) {
 		BZERO(ctx->search_var, sizeof ctx->search_var);
 
-		ctx->isInCirculationModeForConnect	= false;
-		ctx->isInCirculationModeForHelp		= false;
-		ctx->isInCirculationModeForMsg		= false;
-		ctx->isInCirculationModeForNotice	= false;
-		ctx->isInCirculationModeForQuery	= false;
-		ctx->isInCirculationModeForSasl		= false;
-		ctx->isInCirculationModeForSettings	= false;
-		ctx->isInCirculationModeForTheme	= false;
-		ctx->isInCirculationModeForTime		= false;
-		ctx->isInCirculationModeForVersion	= false;
-		ctx->isInCirculationModeForWhois	= false;
-		ctx->isInCirculationModeForZncCmds	= false;
-		ctx->isInCirculationModeForCmds		= false;
-		ctx->isInCirculationModeForChanUsers	= false;
+		ctx->isInCirculationModeForConnect =
+		ctx->isInCirculationModeForCs =
+		ctx->isInCirculationModeForHelp =
+		ctx->isInCirculationModeForMsg =
+		ctx->isInCirculationModeForNotice =
+		ctx->isInCirculationModeForNs =
+		ctx->isInCirculationModeForQuery =
+		ctx->isInCirculationModeForSasl =
+		ctx->isInCirculationModeForSettings =
+		ctx->isInCirculationModeForSquery =
+		ctx->isInCirculationModeForTheme =
+		ctx->isInCirculationModeForTime =
+		ctx->isInCirculationModeForVersion =
+		ctx->isInCirculationModeForWhois =
+		ctx->isInCirculationModeForZncCmds =
+		ctx->isInCirculationModeForCmds =
+		ctx->isInCirculationModeForChanUsers = false;
 
 		if (ctx->matches != NULL)
 			textBuf_destroy(ctx->matches);
@@ -294,9 +323,10 @@ readline_tab_comp_ctx_reset(PTAB_COMPLETION ctx)
 static void
 init_mode_for_connect(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[9]);
+	const char *cp = addrof(ctx->tc->search_var[9]);
 
-	if ((ctx->tc->matches = get_list_of_matching_connect_cmds(p)) == NULL) {
+	if ((ctx->tc->matches = get_list_of_matching_connect_cmds(cp)) ==
+	    NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -306,12 +336,28 @@ init_mode_for_connect(volatile struct readline_session_context *ctx)
 	ctx->tc->isInCirculationModeForConnect = true;
 }
 
+static void
+init_mode_for_cs(volatile struct readline_session_context *ctx,
+    const int offset)
+{
+	const char *cp = addrof(ctx->tc->search_var[offset]);
+
+	if ((ctx->tc->matches = get_list_of_matching_cs_cmds(cp)) == NULL) {
+		output_error("no magic");
+		return;
+	}
+
+	ctx->tc->elmt = textBuf_head(ctx->tc->matches);
+	auto_complete_cs(ctx, ctx->tc->elmt->text);
+	ctx->tc->isInCirculationModeForCs = true;
+}
+
 static void
 init_mode_for_help(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[6]);
+	const char *cp = addrof(ctx->tc->search_var[6]);
 
-	if ((ctx->tc->matches = get_list_of_matching_commands(p)) == NULL) {
+	if ((ctx->tc->matches = get_list_of_matching_commands(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -324,9 +370,9 @@ init_mode_for_help(volatile struct readline_session_context *ctx)
 static void
 init_mode_for_msg(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[5]);
+	const char *cp = addrof(ctx->tc->search_var[5]);
 
-	if ((ctx->tc->matches = get_matches_common(p)) == NULL) {
+	if ((ctx->tc->matches = get_matches_common(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -339,9 +385,9 @@ init_mode_for_msg(volatile struct readline_session_context *ctx)
 static void
 init_mode_for_notice(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[8]);
+	const char *cp = addrof(ctx->tc->search_var[8]);
 
-	if ((ctx->tc->matches = get_matches_common(p)) == NULL) {
+	if ((ctx->tc->matches = get_matches_common(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -351,20 +397,36 @@ init_mode_for_notice(volatile struct readline_session_context *ctx)
 	ctx->tc->isInCirculationModeForNotice = true;
 }
 
+static void
+init_mode_for_ns(volatile struct readline_session_context *ctx,
+    const int offset)
+{
+	const char *cp = addrof(ctx->tc->search_var[offset]);
+
+	if ((ctx->tc->matches = get_list_of_matching_ns_cmds(cp)) == NULL) {
+		output_error("no magic");
+		return;
+	}
+
+	ctx->tc->elmt = textBuf_head(ctx->tc->matches);
+	auto_complete_ns(ctx, ctx->tc->elmt->text);
+	ctx->tc->isInCirculationModeForNs = true;
+}
+
 static void
 init_mode_for_query(volatile struct readline_session_context *ctx)
 {
-	char	*p;
+	const char *cp;
 
 	if (!is_irc_channel(ACTWINLABEL)) {
 		output_error("not in irc channel");
 		return;
 	}
 
-	p = addrof(ctx->tc->search_var[7]);
+	cp = addrof(ctx->tc->search_var[7]);
 
 	if ((ctx->tc->matches = get_list_of_matching_channel_users(ACTWINLABEL,
-	    p)) == NULL) {
+	    cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -377,11 +439,9 @@ init_mode_for_query(volatile struct readline_session_context *ctx)
 static void
 init_mode_for_sasl(volatile struct readline_session_context *ctx)
 {
-	char	*p;
+	const char *cp = addrof(ctx->tc->search_var[6]);
 
-	p = addrof(ctx->tc->search_var[6]);
-
-	if ((ctx->tc->matches = get_list_of_matching_sasl_cmds(p)) == NULL) {
+	if ((ctx->tc->matches = get_list_of_matching_sasl_cmds(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -394,9 +454,9 @@ init_mode_for_sasl(volatile struct readline_session_context *ctx)
 static void
 init_mode_for_set(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[5]);
+	const char *cp = addrof(ctx->tc->search_var[5]);
 
-	if ((ctx->tc->matches = get_list_of_matching_settings(p)) == NULL) {
+	if ((ctx->tc->matches = get_list_of_matching_settings(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -406,14 +466,28 @@ init_mode_for_set(volatile struct readline_session_context *ctx)
 	ctx->tc->isInCirculationModeForSettings = true;
 }
 
+static void
+init_mode_for_squery(volatile struct readline_session_context *ctx)
+{
+	const char *cp = addrof(ctx->tc->search_var[8]);
+
+	if ((ctx->tc->matches = get_list_of_matching_squery_commands(cp)) ==
+	    NULL) {
+		output_error("no magic");
+		return;
+	}
+
+	ctx->tc->elmt = textBuf_head(ctx->tc->matches);
+	auto_complete_squery(ctx, ctx->tc->elmt->text);
+	ctx->tc->isInCirculationModeForSquery = true;
+}
+
 static void
 init_mode_for_theme(volatile struct readline_session_context *ctx)
 {
-	char	*p;
+	const char *cp = addrof(ctx->tc->search_var[7]);
 
-	p = addrof(ctx->tc->search_var[7]);
-
-	if ((ctx->tc->matches = get_list_of_matching_theme_cmds(p)) == NULL) {
+	if ((ctx->tc->matches = get_list_of_matching_theme_cmds(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -426,9 +500,9 @@ init_mode_for_theme(volatile struct readline_session_context *ctx)
 static void
 init_mode_for_time(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[6]);
+	const char *cp = addrof(ctx->tc->search_var[6]);
 
-	if ((ctx->tc->matches = get_matches_common(p)) == NULL) {
+	if ((ctx->tc->matches = get_matches_common(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -441,9 +515,9 @@ init_mode_for_time(volatile struct readline_session_context *ctx)
 static void
 init_mode_for_version(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[9]);
+	const char *cp = addrof(ctx->tc->search_var[9]);
 
-	if ((ctx->tc->matches = get_matches_common(p)) == NULL) {
+	if ((ctx->tc->matches = get_matches_common(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -456,13 +530,13 @@ init_mode_for_version(volatile struct readline_session_context *ctx)
 static void
 init_mode_for_whois(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[7]);
+	const char *cp = addrof(ctx->tc->search_var[7]);
 
 	if (!is_irc_channel(ACTWINLABEL)) {
-		ctx->tc->matches = get_list_of_matching_queries(p);
+		ctx->tc->matches = get_list_of_matching_queries(cp);
 	} else {
 		ctx->tc->matches = get_list_of_matching_channel_users
-		    (ACTWINLABEL, p);
+		    (ACTWINLABEL, cp);
 	}
 
 	if (ctx->tc->matches == NULL) {
@@ -478,9 +552,10 @@ init_mode_for_whois(volatile struct readline_session_context *ctx)
 static void
 init_mode_for_znc_cmds(volatile struct readline_session_context *ctx)
 {
-	char	*p = addrof(ctx->tc->search_var[5]);
+	const char *cp = addrof(ctx->tc->search_var[5]);
 
-	if ((ctx->tc->matches = get_list_of_matching_znc_commands(p)) == NULL) {
+	if ((ctx->tc->matches = get_list_of_matching_znc_commands(cp)) ==
+	    NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -494,10 +569,10 @@ static void
 init_mode_for_commands(volatile struct readline_session_context *ctx,
     const bool numins_greater_than_one)
 {
-	char	*p = addrof(ctx->tc->search_var[1]);
+	const char *cp = addrof(ctx->tc->search_var[1]);
 
 	if (!numins_greater_than_one || (ctx->tc->matches =
-	    get_list_of_matching_commands(p)) == NULL) {
+	    get_list_of_matching_commands(cp)) == NULL) {
 		output_error("no magic");
 		return;
 	}
@@ -535,32 +610,67 @@ no_more_matches(volatile struct readline_session_context *ctx)
 	readline_tab_comp_ctx_reset(ctx->tc);
 }
 
+static bool
+var_matches_cs(CSTRING sv, int *iout)
+{
+	if (strncmp(sv, "/chanserv ", 10) == 0)
+		*iout = 10;
+	else if (strncmp(sv, "/cs ", 4) == 0)
+		*iout = 4;
+	else
+		*iout = 0;
+	return (strncmp(sv, "/chanserv ", 10) == 0 ||
+		strncmp(sv, "/cs ", 4) == 0);
+}
+
+static bool
+var_matches_ns(CSTRING sv, int *iout)
+{
+	if (strncmp(sv, "/nickserv ", 10) == 0)
+		*iout = 10;
+	else if (strncmp(sv, "/ns ", 4) == 0)
+		*iout = 4;
+	else
+		*iout = 0;
+	return (strncmp(sv, "/nickserv ", 10) == 0 ||
+		strncmp(sv, "/ns ", 4) == 0);
+}
+
 static void
 init_mode(volatile struct readline_session_context *ctx)
 {
-	if (!strncmp(get_search_var(ctx), "/connect ", 9))
+	CSTRING sv = get_search_var(ctx);
+	int off1, off2;
+
+	if (!strncmp(sv, "/connect ", 9))
 		init_mode_for_connect(ctx);
-	else if (!strncmp(get_search_var(ctx), "/help ", 6))
+	else if (var_matches_cs(sv, &off1))
+		init_mode_for_cs(ctx, off1);
+	else if (!strncmp(sv, "/help ", 6))
 		init_mode_for_help(ctx);
-	else if (!strncmp(get_search_var(ctx), "/msg ", 5))
+	else if (!strncmp(sv, "/msg ", 5))
 		init_mode_for_msg(ctx);
-	else if (!strncmp(get_search_var(ctx), "/notice ", 8))
+	else if (!strncmp(sv, "/notice ", 8))
 		init_mode_for_notice(ctx);
-	else if (!strncmp(get_search_var(ctx), "/query ", 7))
+	else if (var_matches_ns(sv, &off2))
+		init_mode_for_ns(ctx, off2);
+	else if (!strncmp(sv, "/query ", 7))
 		init_mode_for_query(ctx);
-	else if (!strncmp(get_search_var(ctx), "/sasl ", 6))
+	else if (!strncmp(sv, "/sasl ", 6))
 		init_mode_for_sasl(ctx);
-	else if (!strncmp(get_search_var(ctx), "/set ", 5))
+	else if (!strncmp(sv, "/set ", 5))
 		init_mode_for_set(ctx);
-	else if (!strncmp(get_search_var(ctx), "/theme ", 7))
+	else if (!strncmp(sv, "/squery ", 8))
+		init_mode_for_squery(ctx);
+	else if (!strncmp(sv, "/theme ", 7))
 		init_mode_for_theme(ctx);
-	else if (!strncmp(get_search_var(ctx), "/time ", 6))
+	else if (!strncmp(sv, "/time ", 6))
 		init_mode_for_time(ctx);
-	else if (!strncmp(get_search_var(ctx), "/version ", 9))
+	else if (!strncmp(sv, "/version ", 9))
 		init_mode_for_version(ctx);
-	else if (!strncmp(get_search_var(ctx), "/whois ", 7))
+	else if (!strncmp(sv, "/whois ", 7))
 		init_mode_for_whois(ctx);
-	else if (!strncmp(get_search_var(ctx), "/znc ", 5))
+	else if (!strncmp(sv, "/znc ", 5))
 		init_mode_for_znc_cmds(ctx);
 	else if (ctx->tc->search_var[0] == '/')
 		init_mode_for_commands(ctx, (ctx->numins > 1));
@@ -568,6 +678,22 @@ init_mode(volatile struct readline_session_context *ctx)
 		init_mode_for_channel_users(ctx);
 }
 
+static void
+ac_doit(AC_FUNC ac, volatile struct readline_session_context *ctx)
+{
+	if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
+		no_more_matches(ctx);
+	else
+		ac(ctx, next_text(ctx->tc));
+}
+
+/*
+ * Handles a TAB key press. It initially checks if tabbing is
+ * possible. After that it checks if it's already in circulation for a
+ * mode and auto completes. Else it attempts to store the search
+ * variable and initializes completion for a mode depending on what's
+ * stored in the search variable.
+ */
 void
 readline_handle_tab(volatile struct readline_session_context *ctx)
 {
@@ -577,89 +703,39 @@ readline_handle_tab(volatile struct readline_session_context *ctx)
 		readline_tab_comp_ctx_reset(ctx->tc);
 		return;
 	} else if (ctx->tc->isInCirculationModeForConnect) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_connect(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_connect, ctx);
+	} else if (ctx->tc->isInCirculationModeForCs) {
+		ac_doit(auto_complete_cs, ctx);
 	} else if (ctx->tc->isInCirculationModeForHelp) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_help(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_help, ctx);
 	} else if (ctx->tc->isInCirculationModeForMsg) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_msg(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_msg, ctx);
 	} else if (ctx->tc->isInCirculationModeForNotice) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_notice(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_notice, ctx);
+	} else if (ctx->tc->isInCirculationModeForNs) {
+		ac_doit(auto_complete_ns, ctx);
 	} else if (ctx->tc->isInCirculationModeForQuery) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_query(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_query, ctx);
 	} else if (ctx->tc->isInCirculationModeForSasl) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_sasl(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_sasl, ctx);
 	} else if (ctx->tc->isInCirculationModeForSettings) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_setting(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_setting, ctx);
+	} else if (ctx->tc->isInCirculationModeForSquery) {
+		ac_doit(auto_complete_squery, ctx);
 	} else if (ctx->tc->isInCirculationModeForTheme) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_theme(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_theme, ctx);
 	} else if (ctx->tc->isInCirculationModeForTime) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_time(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_time, ctx);
 	} else if (ctx->tc->isInCirculationModeForVersion) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_version(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_version, ctx);
 	} else if (ctx->tc->isInCirculationModeForWhois) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_whois(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_whois, ctx);
 	} else if (ctx->tc->isInCirculationModeForZncCmds) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_znc_cmd(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_znc_cmd, ctx);
 	} else if (ctx->tc->isInCirculationModeForCmds) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_command(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_command, ctx);
 	} else if (ctx->tc->isInCirculationModeForChanUsers) {
-		if (ctx->tc->elmt == textBuf_tail(ctx->tc->matches))
-			no_more_matches(ctx);
-		else
-			auto_complete_channel_user(ctx, next_text(ctx->tc));
-		return;
+		ac_doit(auto_complete_channel_user, ctx);
 	} else if (store_search_var(ctx) == -1) {
 		output_error("cannot store search variable");
 		return;
